package julia

import (
	"fmt"
	"image"
	"image/color"
	"image/png"
	"math/cmplx"
	"os"
	"sync"

	"github.com/disintegration/imaging"
)

// The function to call on the complex numbers
var fn = DefaultComplexFunction
var clrfn = DefaultColorFunction

// The min and max boundaries to check within
var cMin = complex(-1, -1)
var cMax = complex(1, 1)

// The maximum iterations that the program will take, for vlaues which require
// higher levels of iteration to escape, this should be higher too
var lim = 255

// This contrast value is a percentage between -100 and 100
var contrast = 0.0

// CreateImage will create a png image with the specified resolution
// for a julia function for the function set via JuliaFunction
func CreateImage(filename string, xRes, yRes int) {
	img := image.NewRGBA(image.Rect(0, 0, xRes, yRes))
	workChan := make(chan int, xRes)
	realOffset := real(cMax) - real(cMin)
	imagOffset := imag(cMax) - imag(cMin)

	// Workers function
	var wg sync.WaitGroup
	var worker = func(work chan int) {
		var clr uint8
		step := imagOffset / float64(yRes) // pre-calculate step size
		for w := range work {
			r := real(cMin) + (realOffset / float64(xRes) * float64(w)) // Only need to calculate once per row
			i := imag(cMin)
			for idx := 0; idx < yRes; idx++ {
				clr = calcIter(complex(r, i))
				img.Set(w, idx, clrfn(clr))
				i += step
			}
		}
		wg.Done()
	}

	wg.Add(50)
	for i := 0; i < 50; i++ {
		go worker(workChan)
	}

	// Add the work for the workers
	for x := 0; x < xRes; x++ {
		workChan <- x
	}
	close(workChan)
	wg.Wait()

	newImg := imaging.AdjustContrast(img, contrast)

	writeImage(filename, newImg)
}

// SetColorFunction sets the color of a pixel by using the RGBA value returned
func SetColorFunction(f func(uint8) *color.RGBA) {
	clrfn = f
}

// SetComplexFunction sets the function that will be used to determine the julia
// sets
func SetComplexFunction(f func(complex128, complex128) complex128) {
	fn = f
}

// SetComplexRange sets the range of complex numbers that the image will include
func SetComplexRange(complexMin, complexMax complex128) {
	cMin = complexMin
	cMax = complexMax
}

// SetComplexZoom sets the range of complex numbers by using a coordinate and a zoom amount
// The zoom amount is the range at which the surrounding pixels will be calulated
// So its closer to a radius value
func SetComplexZoom(coord complex128, zoom float64) {
	zoomOffset := zoom / 2
	cMin = complex(real(coord)-zoomOffset, imag(coord)+zoomOffset)
	cMax = complex(real(coord)+zoomOffset, imag(coord)-zoomOffset)
}

// SetMaxIteration sets the maximum iteration count for the fractal
func SetMaxIteration(i int) {
	lim = i
}

// SetContrastValue sets the contrast value for the image, so that more hidden intensities
// can be seen
func SetContrastValue(c float64) {
	contrast = c
}

// DefaultColorFunction is the default function that will color each pixel
// The RGB values of the color will be the iteration count, the alpha is
// always set to 255
func DefaultColorFunction(c uint8) *color.RGBA {
	return &color.RGBA{c, c, c, 255}
}

// DefaultComplexFunction is the function that was used for the daily challenge
// the function is c^2-0.221-0.713i
func DefaultComplexFunction(c, orig complex128) complex128 {
	return c*c + complex(-0.221, -0.713)
}

// calcIter determines how many iterations a complex number took to escape
func calcIter(c complex128) uint8 {
	orig := c
	var iter float64
	for iter = 0; cmplx.Abs(c) < 2 && iter < float64(lim); iter++ {
		c = fn(c, orig)
	}
	iterClamp := (iter / float64(lim)) * 255
	return uint8(iterClamp)
}

// writeImage writes the Image struct to a file using png format
func writeImage(filename string, img image.Image) {
	file, err := os.Create(filename)
	if err != nil {
		fmt.Println("Whoops")
	}
	png.Encode(file, img)
	file.Close()
}
